import os
import json
import gzip
import sqlalchemy
from sqlalchemy.sql.expression import ClauseElement
import operator

from . import models


class Parser:
    def __init__(self, filename="roadTransport.json.gz"):
        self._data = self.readData(filename)
        self._parameters = []

    def readData(self, filename):
        gzip_json = os.path.join(os.path.dirname(__file__), filename)
        if os.path.isfile(gzip_json):
            with gzip.open(gzip_json, "rt") as data_file:
                return json.loads(data_file.read())
        else:
            raise IOError("Can't find file {}".format(filename))

    def _exists(self, model, name, p_name):
        return models.session.query(name(model)).filter_by(name=p_name).scalar() is not None

    @staticmethod
    def get_or_create(session, model, defaults=None, **kwargs):
        instance = session.query(model).filter_by(**kwargs).first()
        if instance:
            return instance, False
        else:
            params = dict((k, v) for k, v in kwargs.items() if not isinstance(v, ClauseElement))
            params.update(defaults or {})
            instance = model(**params)
            session.add(instance)
            return instance, True

    def add(self, parm):
        s = models.session

        # import pdb; pdb.set_trace()
        category, _ = Parser.get_or_create(s, models.Category, name=parm.get("cat"))
        fuel, _ = Parser.get_or_create(s, models.Fuel, name=parm.get("fuel"))
        segment, _ = Parser.get_or_create(s, models.Segment, name=parm.get("segment"))
        eurostd, _ = Parser.get_or_create(s, models.EuroStd, name=parm.get("eurostd"))
        mode, _ = Parser.get_or_create(s, models.Mode, name=parm.get("mode"))
        pollutant, _ = Parser.get_or_create(s, models.Pollutant, name=parm.get("pollutant"))

        # Create parameter instance
        parm_parms = parm.get("parameters")
        parameter = models.Parameter(
            ALPHA=parm_parms.get("Alpha"),
            BETA=parm_parms.get("Beta"),
            DELTA=parm_parms.get("Delta"),
            EPSILON=parm_parms.get("Epsilon"),
            GAMMA=parm_parms.get("Gamma"),
            HTA=parm_parms.get("Hta"),
            MAXSPEED=parm_parms.get("Vmax"),
            MINSPEED=parm_parms.get("Vmin"),
            REDUCTIONFACTOR=parm_parms.get("Reduction Factor [%]"),
            SPEED=parm_parms.get("Speed"),
            ZITA=parm_parms.get("Zita"),
            slope=parm.get("slope"),
            load=parm.get("load"))
        parameter.category = category

        parameter.fuel = fuel
        parameter.segment = segment
        parameter.eurostd = eurostd
        parameter.mode = mode
        parameter.pollutant = pollutant

        # temp store for bulk insertion later
        self._parameters.append(parameter)

    def _parse_data(self):
        if not self._data:
            raise ValueError("No data to parse.. Something went wrong trying to read input data..")

        categories = self._data.get("Type", None)
        if not categories:
            raise AttributeError("Missing 'Type' in JSON file. Inspect file!")

        for c in categories:
            cat_id = c.get("Id")

            fuel = c.get("SSC_NAME")
            for f in fuel:
                fuel_id = f.get("Id")

                subsegments = f.get("Subsegment")
                for s in subsegments:
                    subsegment_id = s.get("Id")

                    euro_standard = s.get("TEC_NAME")
                    for es in euro_standard:
                        es_id = es.get("Id")

                        modes = es.get("Mode")
                        for m in modes:
                            m_id = m.get("Id")

                            slopes = m.get("Slope")
                            for s in slopes:
                                slope_id = s.get("Id")

                                loads = s.get("Load")
                                for l in loads:
                                    l_id = l.get("Id")

                                    pollutants = l.get("Pollutant")
                                    for p in pollutants:
                                        p_id = p.get("Id")

                                        parm = {
                                            "cat": cat_id,
                                            "fuel": fuel_id,
                                            "segment": subsegment_id,
                                            "eurostd": es_id,
                                            "mode": m_id,
                                            "pollutant": p_id,
                                            "slope": slope_id,
                                            "load": l_id,
                                            "parameters": p
                                        }
                                        self.add(parm)

        print("bulk inserting {} parameters".format(len(self._parameters)))
        models.session.add_all(self._parameters)
        models.session.flush()
        models.session.commit()
        print("Done!")



import numpy as np


class Extrapolate(object):
    """Given two input lists representing two axis, this class is able to
    extrapolate the preceding X value
    """

    def __init__(self, x_list, y_list):
        self.xp = x_list
        self.yp = y_list

    def extrap(self, x):
        """np.interp function with linear extrapolation"""
        y = np.interp(x, self.xp, self.yp)
        y = np.where(x < self.xp[0],
                     self.yp[0] + (x - self.xp[0]) * (self.yp[0] - self.yp[1]) / (self.xp[0] - self.xp[1]),
                     y)
        y = np.where(x > self.xp[-1],
                     self.yp[-1] + (x - self.xp[-1]) * (self.yp[-1] - self.yp[-2]) / (self.xp[-1] - self.xp[-2]),
                     y)
        return y

    def __getitem__(self, x):
        return self.extrap(x)



def enum(**named_values):
    return type('Enum', (), named_values)


types = enum(FC='FC', PM='PM', HC='HC', CO='CO', NOx='NOx')


class Pollutants:
    def __init__(self, paths):
        self._pollutants = {}
        self._paths = paths

    def add_pollutant(self, name, is_visible):
        if is_visible:
            self._pollutants[name] = [[] for x in range(self._paths)]

    def __getitem__(self, item):
        return self._pollutants[item]

    def __iter__(self):
        return iter(sorted(self._pollutants))

    def __call__(self):
        return sorted(self._pollutants)

    def __len__(self):
        return len(self._pollutants)



# coding: utf-8
from __future__ import unicode_literals
import sys
import os

from sqlalchemy import Column, Float, Index, Integer, LargeBinary, String, Table, Text, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine, and_, or_
from sqlalchemy.orm import sessionmaker, backref, relationship, scoped_session

# Generate the correct path for the sqlite file
current_path = os.path.dirname(sys.modules[__name__].__file__)
sqlite_file_name = "database.db"
sqlite_full_path = os.path.join(current_path, sqlite_file_name)

Base = declarative_base()
metadata = Base.metadata

# Create the engine connected to the sqlite db
db_engine = create_engine('sqlite:///{}'.format(sqlite_full_path))

# Create the 'Session' class binded with the db engine
Session = scoped_session(sessionmaker(bind=db_engine))

# Create a session instance
session = Session()


def filter_parms(**items):
    """
    To get all segment for 'petrol truck'. Valid parameters:
        cat=category
        fuel=fuel
        segment=segment
        eurostd=eurostd
        pollutant=pollutant
        mode=mode

    >>> truck = session.query(emission.models.Category).all()[2]
    >>> fuel = list(truck.fuels())[0]
    >>> filtered_parameters = emission.models.filter(cat=truck, fuel=fuel)
    >>> set(x.segment for x in filtered_parameters)
    """
    parameters = []
    if "cat" in items:
        parameters = items.get("cat").parameter

    if "fuel" in items:
        parameters = list(filter(lambda x: x.fuel == items.get("fuel"), parameters))

    if "segment" in items:
        parameters = list(filter(lambda x: x.segment == items.get("segment"), parameters))

    if "eurostd" in items:
        parameters = list(filter(lambda x: x.eurostd == items.get("eurostd"), parameters))

    if "pollutant" in items:
        parameters = list(filter(lambda x: x.pollutant == items.get("pollutant"), parameters))

    if "mode" in items:
        parameters = list(filter(lambda x: x.mode == items.get("mode"), parameters))

    return parameters


class Category(Base):
    __tablename__ = 'CATEGORY'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('category'))

    def fuels(self):
        """Find all valid fuel types for this Category vehicle"""
        parm = self.parameter
        vehicle_fuels = set(x.fuel for x in parm)
        return vehicle_fuels

    @classmethod
    def get_for_type(class_, vehicle):
        """Return query to find Category for a particular Vechicle type"""
        Category = class_
        found = session.query(Category).filter_by(name=vehicle.get_category_id()).first()
        return found

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class Fuel(Base):
    __tablename__ = 'FUEL'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('fuel'))

    def segments(self):
        param = self.parameter
        vehicle_segments = set(x.segment for x in param)
        return vehicle_segments

    @classmethod
    def get_for_type(class_, vehicle):
        """Return query to find Fuel for a particluar Vehicle type"""
        Fuel = class_
        found = session.query(Fuel).filter_by(name=vehicle.fuel_type).first()
        return found

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class Segment(Base):
    __tablename__ = 'SEGMENT'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('segment'))

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class EuroStd(Base):
    __tablename__ = 'EUROSTD'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('eurostd'))

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class Pollutant(Base):
    __tablename__ = 'POLLUTANT'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('pollutant'))

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class Mode(Base):
    __tablename__ = 'MODE'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    name = Column('NAME', String, unique=True)
    parameter = relationship('Parameter', backref=backref('mode'))

    def __repr__(self):
        return '{}(name="{}")'.format(self.__class__.__name__, self.name)


class Parameter(Base):
    __tablename__ = 'Parameters'

    ID = Column(Integer, primary_key=True)
    ENT = Column(Integer)
    ALPHA = Column(Float)
    BETA = Column(Float)
    DELTA = Column(Float)
    EPSILON = Column(Float)
    GAMMA = Column(Float)
    HTA = Column(Float)
    MAXSPEED = Column(Float)
    MINSPEED = Column(Float)
    REDUCTIONFACTOR = Column(Float)
    SPEED = Column(Float)
    ZITA = Column(Float)

    slope = Column(Float, nullable=True)
    load = Column(Float, nullable=True)

    category_id = Column(Integer, ForeignKey('CATEGORY.ID'))
    fuel_id = Column(Integer, ForeignKey('FUEL.ID'))
    segment_id = Column(Integer, ForeignKey('SEGMENT.ID'))
    euro_std_id = Column(Integer, ForeignKey('EUROSTD.ID'))
    mode_id = Column(Integer, ForeignKey('MODE.ID'))
    pollutant_id = Column(Integer, ForeignKey('POLLUTANT.ID'))

    @classmethod
    def by_vehicle(class_, vehicle):
        Parameter = class_
        cat = session.query(Category).filter_by(name=vehicle.get_category_id()).first()
        fuel = session.query(Fuel).filter_by(name=vehicle.fuel_type).first()
        query = session.query(Parameter).filter_by(category=cat)
        return query

    def __repr__(self):
        return '{}(id={}, pollutant={}, category={}, fuel={}, segment={}, euro_std={}, mode={}, load={}, slope={})'.format(
            self.__class__.__name__,
            self.ID,
            self.pollutant.name,
            self.category.name,
            self.fuel.name,
            self.segment.name,
            self.eurostd.name,
            self.mode.name,
            self.load,
            self.slope)



import six

def enum(**named_values):
    return type('Enum', (), named_values)


VehicleTypes = enum(CAR='car', BUS='bus', TRUCK='truck', LCATEGORY='L-category', VAN='van')
FuelTypes = enum(PETROL='Petrol',
                 DIESEL='Diesel',
                 PETR_HYB='Petrol Hybrid',
                 LPG_BIFUEL_LPG='LPG Bifuel ~ LPG',
                 LPG_BIFUEL_PETROL='LPG Bifuel ~ Petrol',
                 CNG_BIFUEL_CNG='CNG Bifuel ~ CNG',
                 CNG_BIFUEL_PETROL='CNG Bifuel ~ Petrol',
                 CNG='CNG',
                 BIO_DIESEL='Biodiesel')


class Vehicle(object):
    """Base class for all vehicle types
    """

    mapping = {
        'Passenger Cars': VehicleTypes.CAR,
        'Light Commercial Vehicles': VehicleTypes.VAN,
        'Heavy Duty Trucks': VehicleTypes.TRUCK,
        'Buses': VehicleTypes.BUS,
        'L-Category': VehicleTypes.LCATEGORY
    }

    def __init__(self, vtype, fuel_type, segment, euro_std, mode='', load=-1.0, height=2.0, length=5.0, slope=None):
        self.type = vtype
        self.fuel_type = fuel_type
        self.segment = segment
        self.euro_std = euro_std
        self.mode = mode
        self.load = load
        self.height = height
        self.length = length
        self.slope = slope

    @staticmethod
    def get_type_for_category(category_id):
        return Vehicle.mapping.get(category_id, None)

    def get_category_id(self):
        for cat_id, name in six.iteritems(self.mapping):
            if name == self.type:
                return cat_id


class LCategory(Vehicle):
    """Scooters, 2 and 3-wheels
    """
    def __init__(self, fuel_type=FuelTypes.PETROL, subsegment='', euro_std='', mode='', load=-1.0, height=1.5, length=2.5):
        super(LCategory, self).__init__(VehicleTypes.LCATEGORY, fuel_type, subsegment, euro_std, mode, load, height, length)


class Car(Vehicle):
    """Regular passenger car
        subsegment_id: Mini
        subsegment_id: Small
        subsegment_id: Medium
        subsegment_id: Large-SUV-Executive
    """

    def __init__(self, fuel_type=FuelTypes.PETROL, subsegment='Small', euro_std='Euro 4', mode=''):
        super(Car, self).__init__(VehicleTypes.CAR, fuel_type, subsegment, euro_std, mode)


class Van(Vehicle):
    """Light Commercial Car
        subsegment_id: N1-I
        subsegment_id: N1-II
        subsegment_id: N1-III
    """
    def __init__(self, fuel_type=FuelTypes.DIESEL, subsegment='N1-I', euro_std='Euro 4', mode=''):
        super(Van, self).__init__(VehicleTypes.VAN, fuel_type, subsegment, euro_std, mode)


class Bus(Vehicle):
    """Buses
    subsegment_id: Urban Buses Midi <=15 t
    subsegment_id: Urban Buses Standard 15 - 18 t
    subsegment_id: Urban Buses Articulated >18 t
    subsegment_id: Coaches Standard <=18 t
    subsegment_id: Coaches Articulated >18 t
    """
    def __init__(self, fuel_type=FuelTypes.DIESEL, subsegment='Conventional', euro_std='Euro I', mode='', load=0.0):
        super(Bus, self).__init__(VehicleTypes.BUS, fuel_type, subsegment, euro_std, mode, load)


class Truck(Vehicle):
    """Heavy Duty Trucks
    subsegment_id: Articulated 14 - 20 t
    subsegment_id: Articulated 20 - 28 t
    subsegment_id: Articulated 28 - 34 t
    subsegment_id: Articulated 34 - 40 t
    subsegment_id: Articulated 40 - 50 t
    subsegment_id: Articulated 50 - 60 t
    subsegment_id: Rigid 12 - 14 t
    subsegment_id: Rigid 14 - 20 t
    subsegment_id: Rigid 20 - 26 t
    subsegment_id: Rigid 26 - 28 t
    subsegment_id: Rigid 28 - 32 t
    subsegment_id: Rigid >32 t
    subsegment_id: Rigid 7.5 - 12 t
    subsegment_id: Rigid <=7.5 t
    """
    def __init__(self, fuel_type=FuelTypes.DIESEL, subsegment='Articulated 14 - 20 t', euro_std='Euro I', mode='', load=0.0, slope=0.0):
        super(Truck, self).__init__(VehicleTypes.TRUCK, fuel_type, subsegment, euro_std, mode=mode, load=load, slope=slope)



from bisect import bisect_left


class Interpolate(object):
    def __init__(self, x_list, y_list):
        if any([y - x <= 0 for x, y in zip(x_list, x_list[1:])]):
            raise ValueError("x_list must be in strictly ascending order!")
        x_list = self.x_list = list(map(float, x_list))
        y_list = self.y_list = list(map(float, y_list))
        intervals = zip(x_list, x_list[1:], y_list, y_list[1:])
        self.slopes = [(y2 - y1)/(x2 - x1) for x1, x2, y1, y2 in intervals]

    def __getitem__(self, x):
        if x <= self.x_list[0]:
            return self.y_list[0]
        elif x >= self.x_list[-1]:
            return self.y_list[-1]
        else:
            i = bisect_left(self.x_list, x) - 1
            return self.y_list[i] + self.slopes[i] * (x - self.x_list[i])

if __name__ == "__main__":
    Interpolate([],[])


import os
import json
import gzip
import math

from . import Interpolate
from . import Extrapolate
from . import vehicles
from . import log


class EmissionsJsonParser:
    def __init__(self, vehicle, pollutants, filename="roadTransport.json.gz"):
        self._filename = filename
        self._data = None
        self._parsed_data = {}
        self._vehicle = vehicle

        # TODO: Don't need this?
        self._slope = 0

        self._pollutants = pollutants

        self._read_data()
        self._parse_data()

    def _read_data(self):
        gzip_json = os.path.join(os.path.dirname(__file__), self._filename)
        if os.path.isfile(gzip_json):
            with gzip.open(gzip_json, "rb") as data_file:
                self._data = json.loads(data_file.read())
        else:
            raise IOError("Can't find file {}".format(self._filename))

    @staticmethod
    def get_fuel_type(fuel_id):
        for k, v in vehicles.FuelTypes.__dict__.items():
            if v == fuel_id:
                return v
        return None

    def get_subsegments(self):
        """Util method to list all subsegments for the vehicle
        passed when constructing this object
        """
        subsegments = set()

        categories = self._data.get("Type", None)
        if not categories:
            raise AttributeError("Missing 'Type' in JSON file. Inspect file!")

        for c in categories:
            cat_id = c.get("Id")
            vehicle_type = vehicles.Vehicle.get_type_for_category(cat_id)
            if vehicle_type != self._vehicle.type:
                continue
            # print("cat_id: {}".format(cat_id))

            fuel = c.get("SSC_NAME")
            for f in fuel:
                fuel_id = f.get("Id")
                fuel_type = EmissionsJsonParser.get_fuel_type(fuel_id)

                if not fuel_type:
                    raise ValueError("BAD FUEL TYPE!")

                if fuel_type != self._vehicle.fuel_type:
                    continue

                subsegments = f.get("Subsegment")
                for s in subsegments:
                    subsegment_id = s.get("Id").encode("utf-8")
                    subsegments.add(subsegment_id)
        return subsegments

    def get_euro_standards(self):
        """Util method to list all euro standards for the vehicle
        passed when constructing this object
        """
        euro_standards = set()

        categories = self._data.get("Type", None)
        for c in categories:
            cat_id = c.get("Id")
            vehicle_type = vehicles.Vehicle.get_type_for_category(cat_id)
            if vehicle_type != self._vehicle.type:
                continue
            # print("cat_id: {}".format(cat_id))

            fuel = c.get("SSC_NAME")
            for f in fuel:
                fuel_id = f.get("Id")
                fuel_type = EmissionsJsonParser.get_fuel_type(fuel_id)

                if not fuel_type:
                    raise ValueError("BAD FUEL TYPE!")

                if fuel_type != self._vehicle.fuel_type:
                    continue

                subsegments = f.get("Subsegment")
                for s in subsegments:
                    subsegment_id = s.get("Id").encode("utf-8")

                    if self._vehicle.segment != subsegment_id:
                        continue

                    euro_standard = s.get("TEC_NAME")
                    for es in euro_standard:
                        es_id = es.get("Id")
                        euro_standards.add(es_id)
        return euro_standards

    def get_modes(self):
        modes = set()
        categories = self._data.get("Type", None)
        for c in categories:
            cat_id = c.get("Id")
            vehicle_type = vehicles.Vehicle.get_type_for_category(cat_id)
            if vehicle_type != self._vehicle.type:
                continue

            fuel = c.get("SSC_NAME")
            for f in fuel:
                fuel_id = f.get("Id")
                fuel_type = EmissionsJsonParser.get_fuel_type(fuel_id)

                if not fuel_type:
                    raise ValueError("BAD FUEL TYPE!")

                if fuel_type != self._vehicle.fuel_type:
                    continue

                subsegments = f.get("Subsegment")
                for s in subsegments:
                    subsegment_id = s.get("Id").encode("utf-8")

                    if self._vehicle.segment != subsegment_id:
                        continue

                    euro_standard = s.get("TEC_NAME")
                    for es in euro_standard:
                        es_id = es.get("Id")

                        if self._vehicle.euro_std != es_id:
                            continue

                        modes = es.get("Mode")
                        for m in modes:
                            m_id = m.get("Id")
                            modes.add(m_id)
        return modes

    def _parse_data(self):
        if not self._data:
            raise ValueError("No data to parse.. Something went wrong trying to read input data..")

        categories = self._data.get("Type", None)
        if not categories:
            raise AttributeError("Missing 'Type' in JSON file. Inspect file!")

        mapping = {
            vehicles.VehicleTypes.CAR: vehicles.Car,
            vehicles.VehicleTypes.VAN: vehicles.Van,
            vehicles.VehicleTypes.BUS: vehicles.Bus,
            vehicles.VehicleTypes.LCATEGORY: vehicles.LCategory,
            vehicles.VehicleTypes.TRUCK: vehicles.Truck
        }

        # TODO: Refactore - this was primarily done for testing the JSON structure
        for c in categories:
            cat_id = c.get("Id")
            vehicle_type = vehicles.Vehicle.get_type_for_category(cat_id)
            if vehicle_type != self._vehicle.type:
                continue
            # print("cat_id: {}".format(cat_id))

            fuel = c.get("SSC_NAME")
            for f in fuel:
                fuel_id = f.get("Id")
                fuel_type = EmissionsJsonParser.get_fuel_type(fuel_id)

                if not fuel_type:
                    raise ValueError("BAD FUEL TYPE!")

                if fuel_type != self._vehicle.fuel_type:
                    continue

                subsegments = f.get("Subsegment")
                for s in subsegments:
                    subsegment_id = s.get("Id").encode("utf-8")

                    if self._vehicle.segment != subsegment_id:
                        continue
                    log.debug("subsegment_id: {}".format(subsegment_id))

                    euro_standard = s.get("TEC_NAME")
                    for es in euro_standard:
                        es_id = es.get("Id")

                        if self._vehicle.euro_std != es_id:
                            continue
                        log.debug("es_id: {}".format(es_id))
                        # continue

                        modes = es.get("Mode")
                        for m in modes:
                            m_id = m.get("Id")

                            if self._vehicle.mode != m_id:
                                continue
                            log.debug("mode_id: {}".format(m_id.encode("utf-8")))

                            slopes = m.get("Slope")
                            for s in slopes:
                                slope_id = s.get("Id")

                                #if self._vehicle.slope != slope_id:
                                #    continue
                                log.debug("slope_id: {}".format(slope_id.encode("utf-8")))

                                loads = s.get("Load")
                                for l in loads:
                                    l_id = l.get("Id")

                                    if self._vehicle.load > -1:
                                        if self._vehicle.load != float(l_id):
                                            continue
                                    log.debug("load id: ".format(l_id.encode("utf-8")))

                                    pollutants = l.get("Pollutant")
                                    for p in pollutants:
                                        p_id = p.get("Id")
                                        if p_id in self._pollutants:
                                            new_obj = {
                                                "category": cat_id,
                                                "subsegment": subsegment_id,
                                                "euro_standard": es_id,
                                                "slope": float(slope_id) if slope_id != "" else 0.0,
                                            }
                                            new_obj.update(p)

                                            if not self._pollutants.get(p_id, None):
                                                self._pollutants[p_id] = []
                                            self._pollutants[p_id].append(new_obj)

                                            # print("Pollutant: {}".format(p.get("Id")))
                                            # print("     new_obj: {}".format(new_obj))

    def get_for_pollutant(self, pollutant_id, slope=None):
        if pollutant_id not in self._pollutants:
            raise ValueError("Pollutant ID not in list of pollutations to search for..")

        log.debug("== POLLUTANT_ID = {}".format(pollutant_id))

        pollutant = None
        if len(self._pollutants[pollutant_id]) > 1:
            positive_slopes = [0, 0.02, 0.04, 0.06]
            negative_slopes = [-0.06, -0.04, -0.02, 0]

            # Multiple items in list, meaning we have 
            # various slope values
            x = [x for x in self._pollutants[pollutant_id] if x['slope'] == slope]
            if any(x):
                log.debug("FOUND MATCH: {}".format(slope))
                pollutant = x[0]
                log.debug("      pollutant: {}".format(pollutant))
            else:
                # No match was found. Need to Extrapolate / Interpolate the 
                # emission value
                log.debug("NO MATCH: {}".format(slope))
                slopes_for_pollutant = []
                if slope > 0.0:
                    tmp_pollutants = [x for x in self._pollutants[pollutant_id] if x['slope'] in positive_slopes]
                    slopes_for_pollutant = map(EmissionsJsonParser.calculate, tmp_pollutants)
                    extrapolate = Extrapolate(positive_slopes, slopes_for_pollutant)
                    tmp = extrapolate[slope]
                    log.debug("Extrapolated value: {}".format(tmp))
                    return tmp

                else:
                    tmp_pollutants = [x for x in self._pollutants[pollutant_id] if x['slope'] in negative_slopes]
                    slopes_for_pollutant = map(EmissionsJsonParser.calculate, tmp_pollutants)
                    interpolate = Interpolate(negative_slopes, slopes_for_pollutant)
                    tmp = interpolate[slope]
                    log.debug("Interpolated value: {}".format(tmp))
                    return tmp

        else:
            pollutant = self._pollutants[pollutant_id][0]
        tmp = EmissionsJsonParser.calculate(pollutant)
        log.debug("Regular value: {}".format(tmp))
        return tmp

    @staticmethod
    def calculate(pollutant):
        """
            this calculation is taken from the EU spreadsheet!
        """
        alpha = float(pollutant.get("Alpha"))
        beta = float(pollutant.get("Beta"))
        delta = float(pollutant.get("Delta"))
        epsilon = float(pollutant.get("Epsilon"))
        gamma = float(pollutant.get("Gamma"))
        hta = float(pollutant.get("Hta"))
        reduct_fact = float(pollutant.get("Reduction Factor [%]"))
        speed = float(pollutant.get("Speed"))
        v_max = float(pollutant.get("Vmax"))
        v_min = float(pollutant.get("Vmin"))
        zita = float(pollutant.get("Zita"))

        """ ((alpha*speed^2) + (beta*speed) + gamma + (delta/speed))/((epsilon*speed^2) * (zita * speed + htz))"""
        result = (alpha * math.pow(speed, 2)) + (beta * speed) + gamma + (delta / speed)
        result /= (epsilon * math.pow(speed, 2)) + ((zita * speed) + hta)
        result *= (1 - reduct_fact)
        return result



"""Emission python module

This module is for everyone who wants to build tools / systems for calculating vehicle
emissions for various types of vehicles (Trucks, Buses, Cars, Vans, scooters, ..).

The calculation and factors is provided by the EU EEA guidebook: http://www.eea.europa.eu/publications/emep-eea-guidebook-2016

Main features:
1. Give you the current emission for one or multiple types of pollutants for a given vehicle (check "emission.")
2. Given two points (UTM 33N coordinates - Norway) and a vehicle type, the module will use a 
   routing-service to calculate the best route to reach your destination. The module will also calculate the
   emission for all the pollutant types you have defined. You can then sort the various routes depending of
   which critieria that's most important (duration, pollution (NOx, CO, ..)).

Fuel consumption is directly related to the emission - therefore the lower emission the lower fuel consumption.
For the transportation industry this will be of great importance.
"""

import logging
log = logging.getLogger("emission")

from .Extrapolate import Extrapolate
from .Interpolate import Interpolate
from .Pollutants import Pollutants
from .EmissionJSONReader import EmissionsJsonParser
from .planner import Planner
from .planner import PollutantTypes

from .__version__ import  __version__

__author__ = "NPRA - Norwegian Public Roads Administration"

from .models import session
from . import models
from . import update_db as update


import json
try:
    from urllib.request import urlopen  # Python 3
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlopen  # Python 2
    from urllib import urlencode
import socket
import math

from . import vehicles, log
from . import EmissionsJsonParser
from .exceptions import RouteError
from . import models


def enum(**named_values):
    return type('Enum', (), named_values)


# List of possible pollutant types
PollutantTypes = enum(
    CH4='CH4',
    CO='CO',
    EC='EC',
    NOx='NOx',
    PM_EXHAUST='PM Exhaust',
    VOC='VOC')

# URL to remote route webservice
ROUTE_URL_BASE = "https://www.vegvesen.no/ws/no/vegvesen/ruteplan/routingService_v1_0/routingService/"


class Route:
    """Represent a route object from the NVDB RoutingService"""

    def __init__(self, distance, minutes, path, id):
        self.distance = distance
        self.minutes = minutes
        self.path = path
        self.pollutants = {}
        self.distances = []
        self.id = id

    def hours_and_minutes(self):
        """Return hours:minutes as a string
        representation, based on the total amount
        of minutes for the route.
        """
        hours, minutes = divmod(self.minutes, 60)
        return "{}:{}".format(hours, minutes)

    def velocity(self):
        """Calculate the velocity
        """
        total_time = self.minutes * 60
        return (self.distance / total_time) * 3.6

    def add_pollutant(self, p, calc_emission):
        if p not in self.pollutants:
            self.pollutants[p] = []

        self.pollutants[p].append(calc_emission)

    def add_distances(self, distances):
        self.distances.append(distances)

    def total_emission(self, pollutant):
        total = sum(self.pollutants[pollutant])
        return total

    def __repl__(self):
        fmt = "Route(distance={}, minutes={})"
        return fmt.format(self.distance, self.minutes)

    def __str__(self):
        return self.__repl__()

    def __eq__(self, other):
        return self.minutes == other.minutes

    def __lt__(self, other):
        return self.minutes < other.minutes


class RouteSet:
    """A collection of Route objects"""

    def __init__(self, routes=None):
        if routes is None:
            self._lst = []
        else:
            self._lst = routes

    def __getitem__(self, item):
        return self._lst[item]

    def __iter__(self):
        return iter(self._lst)

    def __len__(self):
        return len(self._lst)

    def add(self, route):
        self._lst.append(route)

    def __repl__(self):
        return "RouteSet({})".format("\n".join([str(r) for r in self._lst]))

    def __str__(self):
        return self.__repl__()

    def sort(self, key=None, reverse=False):
        self._lst.sort(key=key, reverse=reverse)

    def __hash__(self):
        return hash(self._lst)


class Planner:
    """This class takes a start, stop and vehicle input to give the user
    a set of possible road routes sorted after the least pollution. Also
    more metadata about each route is provided.
    """

    def __init__(self, start, stop, vehicle):
        self._start = start
        self._stop = stop

        if not isinstance(vehicle, vehicles.Vehicle):
            raise ValueError("Vehicle is not of correct type. Check vehicle implementations.")
        self._vehicle = vehicle
        # self._emissionJson = EmissionsJsonParser(vehicle)
        # self._emissionJson._init_values_from_input_file()
        self._emissionDb = None  # EmissionsJsonParser(self._vehicle)
        self.routes = RouteSet()

        self._pollutants = {}

    @property
    def pollutants(self):
        return self._pollutants

    def add_pollutant(self, pollutant_type):
        # validate input
        if pollutant_type not in PollutantTypes.__dict__.values():
            raise ValueError("pollutant_type needs to be one of the types defined in planner.PollutantTypes")

        if pollutant_type not in self._pollutants:
            self._pollutants[pollutant_type] = None
        else:
            log.debug("warning: pollutant already added..")
        log.debug("self._pollutants = {}".format(self._pollutants))

    @property
    def coordinates(self):
        return "{start[0]},{start[1]};{end[0]},{end[1]}".format(
            start=self._start, end=self._stop)

    @staticmethod
    def build_url(vehicle, coordinates, format="json", geometryformat="isoz"):
        """Construct a well formed url for the routing service which
        NPRA is using.
        """
        load = vehicle.load if vehicle.load > -1.0 else 0
        params = {
            "format": format,
            "height": vehicle.height,
            "length": vehicle.length,
            "stops": coordinates,
            "load": load,
            "geometryformat": geometryformat,
            "lang": "nb-no",
        }

        return '?'.join([ROUTE_URL_BASE, urlencode(params)])

    def _get_routes(self):
        socket.setdefaulttimeout(30)
        try:
            url = Planner.build_url(self._vehicle, self.coordinates)
            log.debug("Calling: {}".format(url))
            log.debug("coordinates: {}".format(self.coordinates))
            response = urlopen(url)
            data = response.read()
            self._json_data = json.loads(data.decode("utf-8"))
            if 'messages' in self._json_data:
                raise RouteError("Missing 'messages' in returned JSON data.")

        except IOError as err:
            log.debug("ioerror: {}".format(err))
            self._json_data = {}
            raise RouteError("IOError: {}".format(err))

        except ValueError:
            log.warning("Bad data from remote routing service: \n{}".format(data))
            self._json_data = {}
            raise RouteError("Bad data from remote routing service: \n{}".format(data))

    @staticmethod
    def _get_distance_2d(point1, point2):
        distance = math.sqrt((point1[0] - point2[0])**2 + (point1[1] - point2[1])**2)
        return distance

    @staticmethod
    def _get_distance_3d(point1, point2):
        distance = math.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2 + (point2[2] - point1[2]) ** 2)
        return distance

    @staticmethod
    def _get_slope(point1, point2):
        distance = Planner._get_distance_3d(point1, point2)
        slope = 0.0
        if distance:
            slope = math.degrees(math.asin((float(point2[2]) - float(point1[2])) / distance))
        return slope

    def _get_pollutants_for_vehicle(self):
        """Retrieve pollutions parameters for the vehicle provided
        to the planner. Only include the pollutants provided in
        'self._pollutants'
        """
        from . import session

        category = models.Category.get_for_type(self._vehicle)
        if not category:
            raise ValueError("Unable to find Category for vehicle: {}".format(category))

        fuel = session.query(models.Fuel).filter_by(name=self._vehicle.fuel_type).first()
        if not fuel:
            raise ValueError("Unable to find Fuel in database: name={}".format(self._vehicle.fuel_type))

        segment = session.query(models.Segment).filter_by(name=self._vehicle.segment).first()
        if not segment:
            raise ValueError("Unable to find segment in database: name={}".format(str(self._vehicle.segment)))

        filter_parms = {
            "category": category,
            "fuel": fuel,
            "segment": segment
        }

        euro_std = session.query(models.EuroStd).filter_by(name=self._vehicle.euro_std).first()
        if euro_std:
            filter_parms.update({"eurostd": euro_std})

        mode = session.query(models.Mode).filter_by(name=self._vehicle.mode).first()
        if mode:
            filter_parms.update({"mode": mode})

        if self._vehicle.load > -1.0:
            filter_parms.update({"load": self._vehicle.load})

        # Get Parameters based on the other items found above
        params = session.query(models.Parameter).filter_by(**filter_parms)
        return params.all()

    def get_emission(self, parameters, slope=None):
        pollutant = None

        if len(parameters) > 1:
            # We have many parameters instances for a single pollutant.
            # This means that we have multiple 'slopes' in our table.
            # Need therefore to find slope or extrapolate/interpolate the value.
            positive_slopes = [0, 0.02, 0.04, 0.06]
            negative_slopes = [-0.06, -0.04, -0.02, 0]

            x = [x for x in parameters if x.slope == slope]
            if any(x):
                pollutant = x[0]
            else:
                slopes_for_pollutant = []
                if slope > 0.0:
                    tmp_pollutants = [x for x in parameters if x.slope in positive_slopes]
                    slopes_for_pollutant = map(Planner.calculate, tmp_pollutants)
                    extrapolate = Extrapolate(positive_slopes, slopes_for_pollutant)
                    tmp = extrapolate[slope]
                    log.debug("Extrapolated value: {}".format(tmp))
                    return tmp

                else:
                    tmp_pollutants = [x for x in parameters if x.slope in negative_slopes]
                    slopes_for_pollutant = map(Planner.calculate, tmp_pollutants)
                    interpolate = Interpolate(negative_slopes, slopes_for_pollutant)
                    tmp = interpolate[slope]
                    log.debug("Interpolated value: {}".format(tmp))
                    return tmp
        else:
            pollutant = parameters[0]
        tmp = Planner.calculate(pollutant)
        log.debug("tmp: {}".format(tmp))
        return tmp

    @staticmethod
    def calculate(parameter):
        """Equation copied from the EU spreadsheet
        """
        alpha = parameter.ALPHA
        beta = parameter.BETA
        delta = parameter.DELTA
        epsilon = parameter.EPSILON
        gamma = parameter.GAMMA
        hta = parameter.HTA
        reduct_fact = parameter.REDUCTIONFACTOR
        speed = parameter.SPEED
        v_max = parameter.MAXSPEED
        v_min = parameter.MINSPEED
        zita = parameter.ZITA

        """ ((alpha*speed^2) + (beta*speed) + gamma + (delta/speed))/((epsilon*speed^2) * (zita * speed + htz))"""
        result = (alpha * math.pow(speed, 2)) + (beta * speed) + gamma + (delta / speed)
        result /= (epsilon * math.pow(speed, 2)) + ((zita * speed) + hta)
        result *= (1 - reduct_fact)
        return result

    def _calculate_emissions(self):
        """Calculate total emission from a route of x,y,z points based on a path between 
        two points (A -> B). https://www.vegvesen.no/vegkart/vegkart/.

        For a simple static emission calculation play with:
            - self._get_pollutants_for_vehicle()
            - Planner.calculate(parameter)
        """
        parameters = self._get_pollutants_for_vehicle()

        self.routes = RouteSet()

        if "routes" not in self._json_data:
            log.debug("Error in returned JSON data from web service.")
            log.debug("data: {}".format(self._json_data))
            return

        # Create a "set" of Routes. The planner web service will
        # return 2-4 routes with different paths.
        for idx, r in enumerate(self._json_data["routes"]["features"]):
            attributes = r.get("attributes")
            route = Route(distance=attributes.get("Total_Meters"),
                          minutes=attributes.get("Total_Minutes"),
                          path=r.get("geometry").get("paths")[0], id = idx)
            self.routes.add(route)

        log.debug("Nr of routes: {}".format(len(self.routes)))
        for i, route in enumerate(self.routes):
            # A list of x,y,z points that all together represents the route
            path_coordinates = route.path
            distances = []

            # Nifty little trick to loop over 'path_coordinates',
            # but keep a reference to the 'prev' item to calculate the
            # distance between them
            iter_points = iter(path_coordinates)
            prev = next(iter_points)
            for point in path_coordinates:
                if not distances:
                    # first point
                    distances.append(Planner._get_distance_3d(prev, point) / 1000)
                else:
                    distances.append(distances[-1] + Planner._get_distance_3d(prev, point) / 1000)

                point_slope = Planner._get_slope(prev, point)

                # Calculate emission for each pollutants the user has asked for
                for p in self._pollutants:
                    parms = [x for x in parameters if x.pollutant.name.startswith(p)]
                    calc_emission = self.get_emission(parms, point_slope)
                    route.add_pollutant(p, calc_emission)

                prev = point

            route.add_distances(distances)

    def run(self):
        """
        Use the input data and send a HTTP request to route planner.
        Construct a 'Routes' object containing all the possible 'Route' objects.

        Also compute the pollution factor for each route based on the 'Route' data and
        the vehicle choosen.
        """

        self._get_routes()
        self._calculate_emissions()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command


# Meta-data
NAME = "emission"
DESCRIPTION = "Emission calculator library"
DESCRIPTION_OLD = """Emission calculator library to calculate total emissions
given a start and stop position. Various possible routes will be presented
sorted after the least pollutant route.

Vehicle types will be: personal cars, busses and trailers of different sizes.

NOTE that this extension only works for Norway, since we are using the
NPRA (Norwegian Public Roads Administration) road database https://www.vegvesen.no/nvdb/apidokumentasjon/
"""
URL = "https://github.com/NPRA/EmissionCalculatorLib"
EMAIL = "asbjorn@fellinghaug.com"
AUTHOR = """Juraj Cirbus <Juraj.Cirbus@norconsult.com>, Tomas Levin <tomas.levin@vegvesen.no>, Asbjørn Alexander Fellinghaug <asbjorn.fellinghaug@webstep.no>"""

REQUIRED = [
    'numpy',
    'SQLAlchemy',
    'six'
]

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.rst'), encoding="utf-8") as f:
    long_description = '\n' + f.read()

about = {}
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)


class PublishCommand(Command):
    """Support setup.py publish
    """

    description = 'Build and publish the package'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold"""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (univeral) distribution...')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine...')
        os.system('twine upload --repository-url https://test.pypi.org/legacy/ dist/*')
        #os.system('twine upload dist/*')
        #sys.exit()

print("Version: {}".format(about['__version__']))


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    keywords="Emission calculation library using the NVDB (Norway)",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=['emission'],
    install_requires=REQUIRED,
    package_data={
        'emission': ['*.json.gz'],
    },
    include_package_data=True,
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Education',
        'Topic :: Office/Business'
    ],
    # $ setup.py publish support
    cmdclass={
        'publish': PublishCommand,
    },
)
